import os
import logging
import boto3
from botocore.client import Config
import gzip
import re
from uuid import uuid4
import json
import query_string
import smart_open.s3
import shutil
import tempfile
import io
from urllib.parse import urlencode

from .__util_versioned_ref import VersionedRef
from .__util_metadata import FIELDS

DISABLE_GZIP = os.environ.get('DISABLE_GZIP')
ENV = os.environ.get('ENV')
AWS_REGION = os.environ.get('AWS_REGION')
MULTIPART_SIZE = 100 * 1024 * 1024 # don't use multipart if file size is smaller than 100MB

def get_kms_key_name(org_slug):
    return f'alias/customer-key-{ENV}-{org_slug}'

def lowerMetadataKeys(metadata):
    m = {}
    for key, value in metadata.items():
        m[key.lower()] = value
    return m

def resolveCustomMetadataAndTags(metadata):
    custom_metadata_str = metadata.get(FIELDS["CUSTOM_METADATA"], "") or ""
    custom_tags_str = metadata.get(FIELDS["CUSTOM_TAGS"], "") or ""
    if not custom_tags_str:
        custom_tags = []
    else:
        custom_tags = custom_tags_str.split(",")

    return {
        'custom_metadata': query_string.parse(custom_metadata_str),
        'custom_tags': custom_tags
    }

def getOrEmptyString(dic, key, default=''):
    val = dic.get(key, default)
    if val is None:
        return default
    return val

class S3FileobjUploader:
    def __init__(self, s3, fileobj, params, options={'disable_gzip': False}):
        self.s3 = s3
        self.fileobj = fileobj
        self.params = params
        self.disable_gzip = options.get('disable_gzip', False)
        self.stream = io.BytesIO()
        self.partCount = 0
        self.multipart = None
        self.parts = []
        if not self.disable_gzip:
            self.compressor = gzip.GzipFile(fileobj=self.stream, mode='w')

    def _uploadPart(self):
        print(f'upload multipart {self.partCount}')
        if self.partCount == 0:
            self.multipart = self.s3.create_multipart_upload(**self.params)
        self.partCount += 1
        self.stream.seek(0)
        part = self.s3.upload_part(
            Body=self.stream,
            Bucket=self.multipart['Bucket'],
            Key=self.multipart['Key'],
            PartNumber=self.partCount,
            UploadId=self.multipart['UploadId'])
        self.parts.append({ **part, 'PartNumber': self.partCount })
        self.stream.seek(0)
        self.stream.truncate()

    def _uploadLastPart(self):
        if self.partCount == 0:
            self.stream.seek(0)
            return self.s3.put_object(
                Body=self.stream,
                **self.params)
        else:
            self._uploadPart()
            parts = []
            for part in self.parts:
                parts.append({
                    'ETag': part['ETag'],
                    'PartNumber': part['PartNumber']
                })
            return self.s3.complete_multipart_upload(
                Bucket=self.multipart['Bucket'],
                Key=self.multipart['Key'],
                UploadId=self.multipart['UploadId'],
                MultipartUpload={ 'Parts': parts })

    def upload(self):
        while True:
            chunk = self.fileobj.read(1024 * 1024)
            if not chunk:
                if not self.disable_gzip:
                    self.compressor.close()
                return self._uploadLastPart()
            if self.disable_gzip:
                self.stream.write(chunk)
            else:
                self.compressor.write(chunk)
            if self.stream.tell() >= MULTIPART_SIZE:
                self._uploadPart()

class Datalake:
    def __init__(self, endpoint):
        if endpoint:
            self.s3 = boto3.client(
                's3', endpoint_url=endpoint,
                aws_access_key_id='123', aws_secret_access_key='abc',
                region_name=AWS_REGION, config=Config(signature_version='s3v4'))
            self.resource_kwargs = { 'endpoint_url': endpoint }
        else:
            self.s3 = boto3.client('s3', region_name=AWS_REGION, config=Config(signature_version='s3v4'))
            self.resource_kwargs = None

    def get_s3_head(self, file):
        bucket = file['bucket']
        file_key = file['fileKey']
        if 'version' in file:
            file_version = file['version']
            head = self.s3.head_object(Bucket=bucket, Key=file_key, VersionId=file_version)
        else:
            head = self.s3.head_object(Bucket = bucket, Key = file_key)

        return head

    def get_file_meta(self, file):
        head = self.get_s3_head(file)
        return lowerMetadataKeys(head.get('Metadata'))

    def read_file(self, file, form='body'):
        bucket = file['bucket']
        file_key = file['fileKey']
        if 'version' in file:
            kwargs = {'VersionId': file['version']}
        else:
            kwargs = {}

        if form == 'body':
            response = self.s3.get_object(Bucket=bucket, Key=file_key, **kwargs)
        elif form in ['file_obj', 'download']:
            response = self.s3.head_object(Bucket=bucket, Key=file_key, **kwargs)
        else:
            raise ValueError(f'Invalid form={form}; supported values are body, file_obj and download')

        status_code = response.get('ResponseMetadata', {}).get('HTTPStatusCode')
        if not status_code == 200:
            logging.error(response)
            raise Exception('Invalid response code')

        meta = lowerMetadataKeys(response.get('Metadata'))
        result = { 'metadata': meta, **resolveCustomMetadataAndTags(meta) }

        if form == 'body':
            content = response.get('Body').read()
            if response.get('ContentEncoding') == 'gzip':
                result['body'] = gzip.decompress(content)
            else:
                result['body'] = content
        else:
            file_obj = smart_open.s3.open(bucket, file_key, 'rb', file.get('version'), resource_kwargs=self.resource_kwargs)
            if response.get('ContentEncoding') == 'gzip':
                file_obj = gzip.open(file_obj)

            if form == 'download':
                with tempfile.NamedTemporaryFile(mode='wb', delete=False) as tf:
                    shutil.copyfileobj(file_obj, tf)
                    result['download'] = tf.name
                file_obj.close()
            else:
                result['file_obj'] = file_obj

        return result

    def write_file(self, context, content, file_name, file_category, raw_file, file_meta, ids = None, source_type = None):
        bucket = raw_file['bucket']
        raw_file_key = raw_file['fileKey']
        if file_category == 'IDS' and ids is None:
            raise Exception(f'ids can not be None when file_category is IDS')
        ids_obj = VersionedRef(composite=ids)
        pattern = '(.*?)/(.*?)/(?:.*?)/(.*)'
        match = re.match(pattern, raw_file_key)
        if not match:
            raise Exception(f'Raw file key {raw_file_key} does not match "{pattern}"')

        if source_type is not None:
            source_type_match = re.match('^[-a-z0-9]+$', source_type)
            if not source_type_match:
                raise Exception(f'Source type "{source_type}" contains invalid character or upper case letter')
        else:
            source_type = getOrEmptyString(file_meta, FIELDS['SOURCE_TYPE'], 'unknown')

        org_slug, source_id, raw_file_path = match.groups()
        file_key = os.path.join(org_slug, source_id, file_category, raw_file_path, file_name)
        file_id = str(uuid4())
        pipelineConfig = context.get('pipelineConfig', {})
        meta = {
            # constant
            FIELDS['INTEGRATION_TYPE']: 'datapipeline',
            FIELDS['VERSION']: '2',

            # generated
            FIELDS['FILE_ID']: file_id,

            # from raw file
            FIELDS['RAW_FILE_ID']: getOrEmptyString(file_meta, FIELDS['FILE_ID']),
            FIELDS['CUSTOM_METADATA']: getOrEmptyString(file_meta, FIELDS['CUSTOM_METADATA']),
            FIELDS['CUSTOM_TAGS']: getOrEmptyString(file_meta, FIELDS['CUSTOM_TAGS']),
            FIELDS['SOURCE_NAME']: getOrEmptyString(file_meta, FIELDS['SOURCE_NAME']),
            FIELDS['SOURCE_TYPE']: source_type,
            FIELDS['TRACE_ID']: getOrEmptyString(file_meta, FIELDS['TRACE_ID']),

            # IDS/TMP
            **({
                FIELDS['IDS']: ids_obj.composite,
                FIELDS['IDS_TYPE']: ids_obj.name,
                FIELDS['IDS_VERSION']: ids_obj.version,
            } if ids is not None and (file_category == 'IDS' or file_category == 'TMP') else {}),

            # from pipeline context
            FIELDS['INTEGRATION_ID']: getOrEmptyString(context, 'pipelineId'), # pipeline id
            # https://github.com/tetrascience/ts-service-pipeline/blob/development/src/models/create-workflow-command.js#L171
            FIELDS['INTEGRATION_NAME']: getOrEmptyString(pipelineConfig, 'pipelineName'),
            FIELDS['PIPELINE_ID']: getOrEmptyString(context, 'pipelineId'),
            FIELDS['PIPELINE_WORKFLOW_ID']: getOrEmptyString(context, 'workflowId'),
            FIELDS['PIPELINE_MASTER_SCRIPT']: f"{context.get('masterScriptNamespace', '')}/{context.get('masterScriptSlug', '')}:{context.get('masterScriptVersion', '')}",
            FIELDS['PIPELINE_TASK_EXECUTION_ID']: getOrEmptyString(context, 'taskId'),
            FIELDS['PIPELINE_TASK_SCRIPT']: getOrEmptyString(context, 'taskScript'),
            FIELDS['PIPELINE_TASK_SLUG']: getOrEmptyString(context, 'taskSlug')
        }

        params = {
            'Bucket': bucket,
            'Key': file_key,
            'Metadata': meta,
            'ServerSideEncryption': 'aws:kms',
            'SSEKMSKeyId': get_kms_key_name(org_slug),
            'ContentEncoding': 'gzip'
        }

        if hasattr(content, 'read'):
            response = S3FileobjUploader(
                self.s3, content, params, { 'disable_gzip': DISABLE_GZIP }).upload()
        else:
            if not DISABLE_GZIP:
                if (isinstance(content, str)):
                    content = content.encode()
                content = gzip.compress(content)
            response = self.s3.put_object(Body=content, **params)
        logging.info(response)
        return {
            'type': 's3file',
            'bucket': bucket,
            'fileKey': file_key,
            'fileId': file_id,
            # fakeS3 does not return VersionId, so use '' to avoid an exception
            'version': response.get('VersionId', '')
        }

    def update_metadata_tags(self, file, custom_meta, custom_tags):
        bucket = file['bucket']
        file_key = file['fileKey']

        head = self.get_s3_head(file)
        current_meta = lowerMetadataKeys(head.get('Metadata'))

        if not FIELDS['FILE_ID'] in current_meta:
            raise Exception('no FILE_ID in meta!')

        if (not custom_meta) and (not custom_tags):
            raise Exception('No metadata or tags provided')

        isASCII = lambda s: s and isinstance(s, str) and bool(re.match(r'^[\x00-\x7F]*$', s))

        custom_meta_str = current_meta.get(FIELDS['CUSTOM_METADATA'], '') or ''
        current_custom_meta = query_string.parse(custom_meta_str)
        if custom_meta:
            custom_meta_merged = {**current_custom_meta, **custom_meta}
            custom_meta_merged = {k:v for k,v in custom_meta_merged.items() if v is not None}
            for k,v in custom_meta_merged.items():
                if not isASCII(k):
                    raise Exception(f'Metadata key {k} contains non-ASCII character')
                if not isASCII(v):
                    raise Exception(f'Metadata value {v} contains non-ASCII character')
            custom_meta_str = urlencode(custom_meta_merged)

        custom_tags_str = current_meta.get(FIELDS['CUSTOM_TAGS'], '')
        if custom_tags:
            for t in custom_tags:
                if not isASCII(t):
                    raise Exception(f'Tag {t} contains non-ASCII character')
            new_custom_tags = list(set(custom_tags_str.split(',') + custom_tags))
            new_custom_tags.sort()
            custom_tags_str = ','.join(new_custom_tags)

        if len(custom_meta_str) + len(custom_tags_str) >= 1024 * 1.5:
            raise Exception('Metadata and tags length larger than 1.5KB')

        params = {
            'Bucket': bucket,
            'CopySource': f'/{bucket}/{file_key}',
            # 'CopySourceIfUnmodifiedSince': head['LastModified'], # ensure no conflict?
            'Key': file_key,
            'ContentEncoding': head.get('ContentEncoding', None),
            'ContentType': head['ContentType'],
            'Metadata': {
                **current_meta,
                FIELDS['CUSTOM_METADATA']: custom_meta_str,
                FIELDS['CUSTOM_TAGS']: custom_tags_str
            },
            'MetadataDirective': 'REPLACE',
            'ServerSideEncryption': 'aws:kms',
            'SSEKMSKeyId': head.get('SSEKMSKeyId', None),
        }

        params = {k:v for k,v in params.items() if v is not None}
        response = self.s3.copy_object(**params)

        return {
            'type': 's3file',
            'bucket': bucket,
            'fileKey': file_key,
            'fileId': current_meta[FIELDS['FILE_ID']],
            # fakeS3 does not return VersionId, so use '' to avoid an exception
            'version': response.get('VersionId', '')
        }

    def write_ids(self, context, content_obj, file_suffix, raw_file, file_meta, ids, source_type, file_category):
        ids_obj = VersionedRef(composite=ids)
        file_name = f'{ids_obj.name}_{ids_obj.version}_{file_suffix}'
        result = self.write_file(context, json.dumps(content_obj, indent=4), file_name, file_category, raw_file, file_meta, ids, source_type)
        return result

    def get_file_name(self, file):
        file_key = file['fileKey']
        return os.path.basename(file_key)

    def get_presigned_url(self, file, ttl_sec):
        if ttl_sec == None or ttl_sec < 0 or ttl_sec > 900:
            raise Exception(f'Cannot generate pre-signed S3 URL, expiration in seconds must be between 0 and 900, and specified value is {ttl_sec}')

        bucket = file['bucket']
        key = file['fileKey']

        if 'version' in file:
            kwargs = {'VersionId': file['version']}
        else:
            kwargs = {}

        try:
            return self.s3.generate_presigned_url('get_object', Params={
                'Bucket': bucket,
                'Key': key,
                **kwargs
            }, ExpiresIn=ttl_sec)
        except Error as e:
            print(e)

        return None



import os
import sys
from threading import Timer
import traceback
from datetime import datetime, timedelta
from time import sleep, time

from .__task_script_runner import run
from .__util_log import Log
from .__util_thread import ThreadWithTrace
from .__util_task import (extend_task_timeout, poll_task, update_task_status)


def get_run_params(task):
  params = {
    'input': task.get('input'),
    'context_from_arg': task.get('context'),
    'func': task.get('func'),
    'correlation_id': task.get('correlation_id'),
    'func_dir': task.get('func_dir'),
    'store_output': False,
    'storage_type': os.environ.get('TASK_STORAGE_TYPE'),
    'storage_bucket': os.environ.get('TASK_STORAGE_S3FILE_BUCKET'),
    'storage_file_key': os.environ.get('TASK_STORAGE_S3FILE_FILE_KEY'),
    'storage_endpoint': os.environ.get('TASK_STORAGE_S3FILE_ENDPOINT'),
    'artifact_bucket': os.environ.get('ARTIFACT_S3FILE_BUCKET'),
    'artifact_prefix': os.environ.get('ARTIFACT_S3FILE_PREFIX'),
    'artifact_endpoint': os.environ.get('ARTIFACT_S3FILE_ENDPOINT'),
    'artifact_file_key': os.environ.get('ARTIFACT_IDS_SCHEMA_S3FILE_FILE_KEY'),
    'artifact_bucket_private': os.environ.get('ARTIFACT_S3FILE_BUCKET_PRIVATE'),
    'artifact_prefix_private': os.environ.get('ARTIFACT_S3FILE_PREFIX_PRIVATE'),
    'artifact_endpoint_private': os.environ.get('ARTIFACT_S3FILE_ENDPOINT_PRIVATE'),
    'command_endpoint': os.environ.get('COMMAND_ENDPOINT')
  }

  return params

log = Log({})

last_run = {'result': None, 'error': None}
run_state = {'task_worker': None, 'task': None}

def healtcheck_worker():
  task = run_state['task']
  task_worker = run_state['task_worker']

  if task and task_worker:
    task_id = task.get('id')
    try:
      extend_task_timeout(task)
    except Exception as e:
      log.log(f'Error during timeout extension -> killing task {task_id}')
      task_worker.kill()

  Timer(60.0, healtcheck_worker).start()


def task_worker(task):
  run_params = get_run_params(task)
  sys.path.append(run_params.get('func_dir'))
  try:
    last_run['result'] = run(**run_params)
  except Exception as e:
    log.log(log.generate_error(e))
    last_run['error'] = traceback.format_exc()
  sys.path.remove(run_params.get('func_dir'))


if __name__ == '__main__':
  healtcheck_worker()
  
  while True:
    task = poll_task()
    if task:
      task_id = task.get('id')
      log.log(f'Got new task {task_id}')

      last_run['result'] = None
      last_run['error'] = None

      run_state['task'] = task
      task_thread = ThreadWithTrace(name=f'task-{task_id}', target=task_worker, args=(task,))
      run_state['task_worker'] = task_thread
      task_thread.start()
      task_thread.join()

      log.log(f'Task {task_id} thread is completed')

      run_state['task_worker'] = None
      run_state['task'] = None

      if last_run['result'] != None:
        update_task_status(task, last_run['result'])
      else:
        update_task_status(task, {
          'status': 'failed',
          'result': {
            'error': last_run['error'] if last_run['error'] else 'No content returned by worker'
          }
        })



import json
import os
import sys
import re
import time
import traceback
import requests
from datetime import datetime

__type_to_url = {
  'ids': 'ids',
  'master-script': 'master-scripts',
  'task-script': 'task-scripts'
}

__ignore_ssl = False

def __get_env(name: str):
    v = os.environ.get(name)
    if not v:
        raise Exception(f'{name} env is not set!')
    return v

def __get_headers():
    headers = {'x-org-slug': __get_env('TS_ORG')}
    ts_auth = __get_env('TS_AUTH_TOKEN')
    if re.compile(r'^([a-z0-9]+-)+[a-z0-9]+$').match(ts_auth, re.IGNORECASE):
        headers['x-api-key'] = ts_auth
    else:
        headers['ts-auth-token'] = ts_auth
    return headers

def set_ignore_ssl(v: bool):
    global __ignore_ssl
    __ignore_ssl = v

def upload_artifact(cfg, artifact_bytes):
    url = f'{__get_env("TS_API_URL")}/artifact/{__type_to_url[cfg.type]}/{cfg.namespace}/{cfg.slug}/{cfg.version}'
    r = requests.post(
        url, 
        headers=__get_headers(),
        params={'force': 'true'} if cfg.force else {},
        data=artifact_bytes,
        verify=not __ignore_ssl
    )
    if r.status_code < 400:
        return r.json()
    else:
        print(r.json(), file=sys.stderr, flush=True)
        raise Exception(f'HTTP status: {r.status_code}, url: {r.url}')

def get_task_script_build_info(id: str):
    url = f'{__get_env("TS_API_URL")}/artifact/builds/{id}'
    r = requests.get(
        url, 
        headers=__get_headers(),
        verify=not __ignore_ssl
    )
    if r.status_code < 400:
        return r.json()
    else:
        print(r.json(), file=sys.stderr, flush=True)
        raise Exception(f'HTTP status: {r.status_code}, url: {r.url}')

def get_task_script_build_logs(id: str, params):
    url = f'{__get_env("TS_API_URL")}/artifact/build-logs/{id}'
    r = requests.get(
        url, 
        headers=__get_headers(),
        params={k:v for k,v in params.items() if v is not None},
        verify=not __ignore_ssl
        )
    if r.status_code < 400:
        return r.json()
    else:
        print(r.json(), file=sys.stderr, flush=True)
        raise Exception(f'HTTP status: {r.status_code}, url: {r.url}')


from argparse import ArgumentParser

from .__put_cmd import put_cmd_args
from .__init_cmd import init_cmd_args
from .__utils import check_update_required
from .._version import __version__

def main():
    check_update_required(__version__)

    parser = ArgumentParser(prog='ts-sdk')
    subparsers = parser.add_subparsers()

    init_cmd_args(subparsers.add_parser(
        'init', 
        help='initialize master and task script from a template'
        ))

    put_cmd_args(subparsers.add_parser(
        'put', 
        help='puts artifact identified by namespace/slug:version'
        ))

    args = parser.parse_args()
    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_help()

if __name__ == '__main__':
    main()



import io
import tempfile
import typing as t
import typing_extensions as te

from ..task.types import File, FileCategory, ReadResult
from ..task.__util_log import Log


class Context:
    """A development-time version of the context object that is passed into
    the task script handler when running as part of a pipeline.
    """

    def __init__(self, pipeline_config = {}):
        self._storage = {}
        self._pipeline_config = pipeline_config
        self._log = Log({})

    @property
    def pipeline_config(self) -> t.Dict[str, str]:
        """Pipeline configuration including secrets."""
        return self._pipeline_config

    def read_file(self, file: File, form: str = 'body') -> ReadResult:
        if form == 'body':
            return self._storage[file["fileKey"]]
        elif form == 'file_obj':
            result = self._storage[file["fileKey"]]
            body = result['body']
            buf = io.BytesIO()
            buf.write(body)
            buf.seek(0)
            result['file_obj'] = buf
            return result
        elif form == 'download':
            result = self._storage[file["fileKey"]]
            with tempfile.NamedTemporaryFile(mode='wb', delete=False) as tf:
                tf.write(result['body'])
                result['download'] = tf.name
            return result
        raise Exception(f'Invalid form: {form}')

    def write_file(
        self,
        content: t.AnyStr,
        file_name: str,
        file_category: FileCategory,
        ids: str = None,
        custom_metadata: t.Dict[str, str] = None,
        custom_tags: t.List[str] = None,
        source_type: str = None,
    ) -> File:
        if type(content) == str:
            content = content.encode('UTF-8')
        self._storage[file_name] = {
            "metadata": {
                "TS_IDS": ids,
                "TS_SOURCE_TYPE": source_type,
                "TS_FILE_CATEGORY": file_category,
            },
            "body": content,
            "custom_metadata": custom_metadata,
            "custom_tags": custom_tags,
        }
        return {
            "type": "s3",
            "bucket": "fake-unittest-bucket",
            "fileKey": file_name,
        }

    def get_ids(namespace: str, slug: str, version: str):
        return {}

    # always return true in local context
    def validate_ids(
        self,
        data: dict,
        namespace: str,
        slug: str,
        version: str
    ) -> bool:
        return True

    def write_ids(
        self,
        content_obj,
        file_suffix: str,
        ids: t.Optional[str] = None,
        custom_metadata: t.Mapping[str, str] = {},
        custom_tags: t.Iterable[str] = [],
        source_type: t.Optional[str] = None,
        file_category: t.Optional[str] = 'IDS'
    ) -> File:
        return {}

    def get_file_name(self, file: File) -> str:
        return ''

    def get_logger(self):
        return self._log

    def get_secret_config_value(self, secret_name: str, silent_on_error=True) -> str:
        return self._pipeline_config[secret_name]

    def get_presigned_url(self, file: File, ttl_sec=300) -> str:
        return ''

    def update_metadata_tags(
        self,
        file: File,
        custom_meta: t.Mapping[str, str] = {},
        custom_tags: t.Iterable[str] = []
    ) -> File:
        return {}

    def run_command(self, org_slug, target_id, action, metadata, payload, ttl_sec=300):
        return {}



# https://www.geeksforgeeks.org/python-different-ways-to-kill-a-thread
import sys
import trace
import threading
import time


class ThreadWithTrace(threading.Thread):
    def __init__(self, *args, **keywords):
        threading.Thread.__init__(self, *args, **keywords)
        self.killed = False

    def start(self):
        self.__run_backup = self.run
        self.run = self.__run
        threading.Thread.start(self)

    def __run(self):
        sys.settrace(self.globaltrace)
        self.__run_backup()
        self.run = self.__run_backup

    def globaltrace(self, frame, event, arg):
        if event == "call":
            return self.localtrace
        else:
            return None

    def localtrace(self, frame, event, arg):
        if self.killed:
            if event == "line":
                raise SystemExit()
        return self.localtrace

    def kill(self):
        self.killed = True



import sys
import json
import importlib
import argparse
import os
import base64
import re
import typing as t
import typing_extensions as te
import uuid

from .types import FileCategory, File, ReadResult

from .__util_storage import Storage
from .__util_datalake import Datalake
from .__util_log import Log
from .__util_merge import merge_arrays, merge_objects
from .__util_metadata import FIELDS
from .__util_ids import create_ids_util
from .__util_command import Command

COMPLETED = 'completed'
FAILED = 'failed'

LOG_TAG_PRE_FUNCTION = 'pre_function_call'
LOG_TAG_POST_FUNCTION = 'post_function_call'
LOG_TAG_SCRIPT_STARTED = 'script_started'
LOG_TAG_SCRIPT_ENDED = 'script_ended'

if 'default_print' not in __builtins__:
  __builtins__['default_print'] = __builtins__['print']

def wrap_log(func_name):
    def return_wrapper(fn):
        def wrapper(*args, **kwargs):
            id = str(uuid.uuid4())
            Context.log.log({
                'level': 'debug',
                'tag': LOG_TAG_PRE_FUNCTION,
                'funcName': func_name,
                'id': id
            })
            result = fn(*args, **kwargs)
            Context.log.log({
                'level': 'debug',
                'tag': LOG_TAG_POST_FUNCTION,
                'funcName': func_name,
                'id': id
            })
            return result
        return wrapper
    return return_wrapper

def camel_to_snake(str) -> str:
    return ''.join(['_'+i.lower() if i.isupper() else i for i in str]).lstrip('_')

class Context:
    """A context object that is passed into
    the task script handler when running as part of a pipeline.
    """

    org_slug: str
    pipeline_id: str
    workflow_id: str

    master_script_namespace: str
    master_script_slug: str
    master_script_version: str

    input_file: File

    created_at: str
    task_id: str
    task_created_at: str

    pipeline_config: t.Mapping[str, str]

    platform_url: str

    tmp_dir: str = '/tmp'


    def __init__(self, obj, datalake, ids_util, log, command):
        self._obj = { **obj, 'tmpDir': '/tmp' } # keys are later converted to snake case via "camel_to_snake"
        for key in self._obj:
            setattr(self, camel_to_snake(key), self._obj[key])
        self._datalake = datalake
        self._ids_util = ids_util
        self._log = log
        self._command = command

    @wrap_log('context.read_file')
    def read_file(self, file: File, form: str = 'body') -> ReadResult:
        """Reads a file from the data lake and returns its contents in one of
        three forms.

        If form='body' (the default), then result['body'] holds the contents of
        the file as a byte array. This approach cannot handle large files that
        don't fit in memory.

        If form='file_obj', then result['file_obj'] is a file-like object that
        can be used to access the body in a streaming manner. This object can
        be passed to Python libraries such as Pandas.

        If form='download', then result['download'] is the file name of a local
        file that has been downloaded from the specified data lake file. This
        is useful when the data needs to be processed by native code (e.g.
        SQLite) or an external utility program.

        >>> import json
        >>> import pandas as pd
        >>> import sqlite3
        >>> def task(input, context):
        ...     f = context.read_file(input, form='body')
        ...     json.loads(f['body'])
        ...
        ...     f = context.read_file(input, form='file_obj')
        ...     df = pd.read_csv(f['file_obj'])
        ...
        ...     f = context.read_file(input, form='download')
        ...     con = sqlite3.connect(f['download'])
        ...     df = pd.read_sql_query('SELECT * FROM foo', con)
        """

        return self._datalake.read_file(file, form)

    @wrap_log('context.write_file')
    def write_file(
        self,
        content: t.Union[bytes, t.BinaryIO, str],
        file_name: str,
        file_category: FileCategory,
        ids: t.Optional[str] = None,
        custom_metadata: t.Mapping[str, str] = {},
        custom_tags: t.Iterable[str] = [],
        source_type: t.Optional[str] = None,
    ) -> File:
        """Writes an output file to the data lake
        """

        raw_file = self.input_file
        file_meta = {
            # in case custom_metadata & custom_tags are undefined in raw_file meta
            FIELDS['CUSTOM_METADATA']: '',
            FIELDS['CUSTOM_TAGS']: '',
            **self._datalake.get_file_meta(raw_file)
        }
        file_meta[FIELDS['CUSTOM_METADATA']] = merge_objects(
            file_meta.get(FIELDS['CUSTOM_METADATA'], ''), custom_metadata
        )
        file_meta[FIELDS['CUSTOM_TAGS']] = merge_arrays(
            file_meta.get(FIELDS['CUSTOM_TAGS'], ''), custom_tags,
        )
        return self._datalake.write_file(
            context=self._obj,
            content=content,
            file_name=file_name,
            file_category=file_category,
            raw_file=raw_file,
            file_meta=file_meta,
            ids=ids,
            source_type=source_type
        )

    @wrap_log('context.get_ids')
    def get_ids(self, namespace: str, slug: str, version: str):
        """Returns IDS schema
        """
        return self._ids_util['get_ids'](namespace, slug, version)

    @wrap_log('context.validate_ids')
    def validate_ids(self, data, namespace: str, slug: str, version: str):
        """Checks validity of IDS content provided in `data`.
        Throws an error if not valid.
        """
        return self._ids_util['validate_ids'](data, namespace, slug, version)

    @wrap_log('context.write_ids')
    def write_ids(
        self,
        content_obj,
        file_suffix: str,
        ids: t.Optional[str] = None,
        custom_metadata: t.Mapping[str, str] = {},
        custom_tags: t.Iterable[str] = [],
        source_type: t.Optional[str] = None,
        file_category: t.Optional[str] = 'IDS'
    ) -> File:
        """Similar to write_file, but for IDS
        """
        raw_file = self.input_file
        file_meta = {
            # in case custom_metadata & custom_tags are undefined in raw_file meta
            FIELDS['CUSTOM_METADATA']: '',
            FIELDS['CUSTOM_TAGS']: '',
            **self._datalake.get_file_meta(raw_file)
        }
        file_meta[FIELDS['CUSTOM_METADATA']] = merge_objects(
            file_meta.get(FIELDS['CUSTOM_METADATA'], ''), custom_metadata
        )
        file_meta[FIELDS['CUSTOM_TAGS']] = merge_arrays(
            file_meta.get(FIELDS['CUSTOM_TAGS'], ''), custom_tags,
        )
        if file_category != 'IDS' and file_category != 'TMP':
            file_category = 'IDS'
        return self._datalake.write_ids(
            context=self._obj,
            content_obj=content_obj,
            file_suffix=file_suffix,
            raw_file=raw_file,
            file_meta=file_meta,
            ids=ids,
            source_type=source_type,
            file_category=file_category
        )

    def get_file_name(self, file: File) -> str:
        """Returns the filename of the file not downloading it locally
        """
        return self._datalake.get_file_name(file)

    def get_logger(self):
        """Returns the structured logger object.
        The input should be an object (eg. containing a message field, among others).

        ...     logger = context.get_logger()
        ...     logger.log({
        ...         "message": "Starting the main parser",
        ...         "level": "info"
        ...     })

        """
        return self._log

    def get_secret_config_value(self, secret_name: str, silent_on_error=True) -> str:
        """Returns the value of the secret.
        If secret is missing, empty string or throws error, depending on the second argument.
        """
        return get_secret_config_value(self._obj, secret_name, silent_on_error)

    def get_presigned_url(self, file: File, ttl_sec=300) -> str:
        """Returns a time-limited HTTPS URL that can be used to access the file.
        If URL generation fails for any reason (except invalid value for ttl_sec parameter) `None` will be returned.
        """
        return self._datalake.get_presigned_url(file, ttl_sec)

    @wrap_log('context.update_metadata_tags')
    def update_metadata_tags(
        self,
        file: File,
        custom_meta: t.Mapping[str, str] = {},
        custom_tags: t.Iterable[str] = []
    ) -> File:
        """Updates file's custom metadata and tags.
        Use 'None' to remove a meta entry.
        New tags will be appended to existing ones.
        """
        return self._datalake.update_metadata_tags(file, custom_meta, custom_tags)

    def run_command(self, org_slug, target_id, action, metadata, payload, ttl_sec=300):
        """Invokes remote command/action on target (agent or connector) and returns its response
        """
        return self._command.run_command(self._obj, org_slug, target_id, action, metadata, payload, ttl_sec)

def output_response(storage, response, correlation_id):
    storage.writeObject({**response, 'id': correlation_id})

def resolve_func(func_dir, func_slug):
    func_conf_file = os.path.join(func_dir, 'config.json')
    with open(func_conf_file, 'r') as file:
        func_conf = json.load(file)
    for f in func_conf['functions']:
        if (f['slug'] == func_slug):
            function = f['function']
            break
    else:
        raise Exception(f'function not found: {func_slug}')
    # print(function)
    func_module, _, func_name = function.rpartition('.')
    return func_module, func_name

def resolve_secrets_in_pipeline_config(context_from_arg):
    secrets = {}
    pipeline_config = context_from_arg.get('pipelineConfig')

    for key in pipeline_config:
      if key.startswith('ts_secret_name_'):
        secret_name = key.split("ts_secret_name_", 1)[1]
        secrets[secret_name] = get_secret_config_value(context_from_arg, key, True)

    return secrets

def get_secret_config_value(context_from_arg, secret_name, silent_on_error=True):
  pipeline_config = context_from_arg.get('pipelineConfig')

  if secret_name.startswith('ts_secret_name_'):
    secret_short_name = secret_name.split("ts_secret_name_", 1)[1]
    secret_full_key = secret_name
  else:
    secret_short_name = secret_name
    secret_full_key = 'ts_secret_name_' + secret_name

  if not secret_full_key in pipeline_config.keys() or pipeline_config.get(secret_full_key) is None:
    if silent_on_error:
      Context.log.log(f'Secret {secret_full_key} not found in the workflow config.')
      return ''
    else:
      raise Exception(f'Secret {secret_full_key} not found in the workflow config.')

  secret_ssm_name = pipeline_config.get(secret_full_key)

  # TODO: remove EKS related code!

  k8s_secret_name_parts = secret_ssm_name.split('/')
  k8s_secret_name = k8s_secret_name_parts[len(k8s_secret_name_parts) - 1]

  try:
    if os.environ.get('TASK_SCRIPTS_CONTAINERS_MODE') == 'ecs':
      secret_value = os.environ.get('TS_SECRET_' + re.sub(r"[^a-z0-9]+", '_', secret_short_name, flags=re.IGNORECASE))
      if secret_value == None:
        raise Exception(f'Secret {secret_short_name} not found')
      return secret_value

    # legacy EKS: read from mounted volume
    secret_value_file = open(f'/etc/ts-secrets-volume/{k8s_secret_name}', 'r')
    secret_value = secret_value_file.read()
    secret_value_file.close()

    return secret_value
  except Exception as e:
    if silent_on_error:
      Context.log.log(e)
      return ''
    else:
      raise Exception(f'Could not resolve secret value for  {secret_full_key}') from e

def run(input, context_from_arg, func, correlation_id, func_dir,
        storage_type, storage_bucket, storage_file_key, storage_endpoint,
        artifact_bucket, artifact_prefix, artifact_endpoint, artifact_file_key,
        artifact_bucket_private, artifact_prefix_private, artifact_endpoint_private, command_endpoint,
        store_output=True):
    log = Log(context_from_arg)
    Context.log = log
    log.log({ 'level': 'debug', 'tag': LOG_TAG_SCRIPT_STARTED })
    if (storage_type != 's3file'):
        raise Exception(f'Invalid storage type: {storage_type}')

    context_from_arg['pipelineConfig'] = {
      **context_from_arg.get('pipelineConfig'),
      **resolve_secrets_in_pipeline_config(context_from_arg)
    }

    # override print function with our own, which decorates with workflow id and task id
    __builtins__['print'] = lambda *args, flush=False: log.log(*args)

    prefix = artifact_prefix if artifact_prefix else ''
    if not prefix and artifact_file_key:
        prefix = re.sub(r"ids/.*$", '', artifact_file_key)
    ids_util = create_ids_util([
        {
            'bucket': artifact_bucket,
            'prefix': prefix,
            'endpoint': artifact_endpoint,
            'namespacePattern': r"^(common|client-.*)$"
        },
        {
            'name': 'private',
            'bucket': artifact_bucket_private,
            'prefix': artifact_prefix_private if artifact_prefix_private else '',
            'endpoint': artifact_endpoint_private,
            'namespacePattern': r"^(private-.*)$"
        }
    ])
    storage = Storage(storage_bucket, storage_file_key, storage_endpoint)
    datalake = Datalake(storage_endpoint)
    command = Command(command_endpoint)
    context = Context(context_from_arg, datalake, ids_util, log, command)
    func_module, func_name = resolve_func(func_dir, func)
    try:
        log.log({ 'level': 'debug', 'tag': LOG_TAG_PRE_FUNCTION, 'funcName': f'{func_module}.{func_name}' })
        result = getattr(importlib.import_module(func_module), func_name)(input, context)
        log.log({ 'level': 'debug', 'tag': LOG_TAG_POST_FUNCTION, 'funcName': func_name })

        if store_output:
            output_response(storage, { 'type': COMPLETED, 'result': result }, correlation_id)

        taskResult = {
          'status': 'completed',
          'result': result or {}
        }
    except Exception as e:
        log.log(e)

        if store_output:
            output_response(storage, { 'type': FAILED, 'error': str(e) }, correlation_id)

        taskResult = {
          'status': 'failed',
          'result': {
            'error': str(e)
          }
        }
    log.log({ 'level': 'debug', 'tag': LOG_TAG_SCRIPT_ENDED })
    return taskResult



import importlib
import sys
import pathlib
import json


def main(func_dir):
    sys.path.append(func_dir)
    with (pathlib.Path(func_dir) / "config.json").open() as f:
        func_conf = json.load(f)
    slugs = set()
    for func in func_conf["functions"]:
        slug = func["slug"]
        if slug in slugs:
            sys.exit(f"Duplicate function slug {slug}")
        slugs.add(slug)

        func_module, _, func_name = func["function"].rpartition(".")
        if not hasattr(importlib.import_module(func_module), func_name):
            sys.exit(f"Unable to resolve function {func_name} in module {func_module}")


if __name__ == "__main__":
    main(sys.argv[1])



# These should be kept in sync with S3MetadataFields.js in ts-lib-shared-schema
FIELDS = {
    'VERSION': 'ts_version',
    'INTEGRATION_TYPE': 'ts_integration_type',
    'INTEGRATION_NAME': 'ts_integration_name',
    'INTEGRATION_ID': 'ts_integration_id', # TODO new?

    'FILE_ID': 'ts_file_id',
    'FILE_TYPE': 'ts_processed_file_type',
    'FILE_PATH': 'ts_file_path', # how to specify this?
    'FILE_NAME': 'ts_file_name',

    'SOURCE_TYPE': 'ts_source_type',
    'SOURCE_NAME': 'ts_source_name',

    'RAW_FILE_ID': 'ts_source_file_id',
    'RAW_FILE_VERSION': 'ts_raw_file_version',

    'IDS': 'ts_ids',
    'IDS_NAMESPACE': 'ts_ids_namespace',
    'IDS_TYPE': 'ts_ids_type',
    'IDS_VERSION': 'ts_ids_type_version',

    'CUSTOM_METADATA': 'ts_integration_metadata',
    'CUSTOM_TAGS': 'ts_integration_tags',

    'PIPELINE_ID': 'ts_pipeline_id',
    'PIPELINE_MASTER_SCRIPT': 'ts_master_script',
    'PIPELINE_TASK_SCRIPT': 'ts_task_script',
    'PIPELINE_TASK_SLUG': 'ts_task_slug',
    'PIPELINE_TASK_EXECUTION_ID': 'ts_task_execution_id',
    'PIPELINE_WORKFLOW_ID': 'ts_workflow_id',

    'TRACE_ID': 'ts_trace_id'
}



import sys
import json
import importlib
import argparse
import os
import base64

from .__task_script_runner import run

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--func', help='function slug', required=True)
    parser.add_argument('--correlation-id', help='correlation id')
    parser.add_argument('--input', help='input string', default='')
    parser.add_argument('--context', help='context', default='')
    parser.add_argument('--func-dir', help='function dir', default='./func')
    args = parser.parse_args()
    params = {
        'input': json.loads(base64.standard_b64decode(args.input)),
        'context_from_arg': json.loads(base64.standard_b64decode(args.context)),
        'func': args.func,
        'correlation_id': args.correlation_id,
        'func_dir': args.func_dir,
        'storage_type': os.environ.get('TASK_STORAGE_TYPE'),
        'storage_bucket': os.environ.get('TASK_STORAGE_S3FILE_BUCKET'),
        'storage_file_key': os.environ.get('TASK_STORAGE_S3FILE_FILE_KEY'),
        'storage_endpoint': os.environ.get('TASK_STORAGE_S3FILE_ENDPOINT'),
        'artifact_bucket': os.environ.get('ARTIFACT_S3FILE_BUCKET'),
        'artifact_prefix': os.environ.get('ARTIFACT_S3FILE_PREFIX'),
        'artifact_endpoint': os.environ.get('ARTIFACT_S3FILE_ENDPOINT'),
        'artifact_file_key': os.environ.get('ARTIFACT_IDS_SCHEMA_S3FILE_FILE_KEY'),
        'artifact_bucket_private': os.environ.get('ARTIFACT_S3FILE_BUCKET_PRIVATE'),
        'artifact_prefix_private': os.environ.get('ARTIFACT_S3FILE_PREFIX_PRIVATE'),
        'artifact_endpoint_private': os.environ.get('ARTIFACT_S3FILE_ENDPOINT_PRIVATE'),
        'command_endpoint': os.environ.get('COMMAND_ENDPOINT')
    }
    sys.path.append(params['func_dir'])

    run(**params)



from importlib.resources import read_text
import json
import typing as t

def load(name: str) -> t.Any:
    return json.loads(read_text(__package__, name))

config = load('config.schema.json')

protocol = load('protocol.schema.json')

__all__ = ['config', 'pipeline']



import typing as t
import os

from .context import Context
from ..task.types import File

def load_pipeline_config(context: Context, config: t.Dict[str, str]):
    """Load config into this context's pipeline config."""
    context._pipeline_config = config


def load_test_file(
    context: Context, local_path: os.PathLike, datalake_path: str, **kwargs
) -> File:
    """Load a local file into the testing data lake, for use in an
    integration test.

    >>> load_test_file(ctx, base_dir / 'setup.py', '/setup.py', file_category='RAW')
    {'type': 's3', 'bucket': 'fake-unittest-bucket', 'fileKey': '/setup.py'}
    """
    with open(local_path) as f:
        return context.write_file(f.read(), datalake_path, **kwargs)


def check_test_output(
    context: Context,
    local_path: os.PathLike,
    datalake_file: File,
    transform: t.Callable[[bytes], t.Any] = None,
):
    """Assert that a file in the testing data lake matches the contents of a
    local file. Optionally transforms the file contents before comparing
    (for example, to ignore whitespace and text ordering for JSON files).

    >>> check_test_output(ctx, base_dir / 'expected.json', output_file, json.loads)
    """
    with open(local_path, "rb") as f:
        expected = f.read()
    actual = context.read_file(datalake_file)["body"]
    if transform is not None:
        expected = transform(expected)
        actual = transform(actual)
    assert expected == actual



import os
from pathlib import Path
import time
import zipfile
import requests
import xml.etree.ElementTree as ET

def get_latest_version():
    try:
        r = requests.get('https://pypi.org/rss/project/ts-sdk/releases.xml', timeout=5)
        root = ET.fromstring(r.content)
        return root.findall('channel/item/title')[0].text
    except:
        return '0.0.0'

def check_update_required(current_version):
    try:
        latest_version_path = Path.home() / '.ts-sdk.latest'
        latest_version = '0.0.0'

        # refresh saved latest version once per day
        if latest_version_path.is_file() and time.time() - latest_version_path.stat().st_mtime < 24 * 3600:
            latest_version = latest_version_path.read_text()
        else:
            latest_version = get_latest_version()
            latest_version_path.write_text(latest_version)

        if latest_version and check_versions_for_update(current_version, latest_version):
            print(f'\n\u001b[33mPlease upgrade ts-sdk (local: {current_version}, latest: {latest_version})\u001b[0m')
            print('\u001b[33mUse: pip3 install ts-sdk --upgrade\u001b[0m\n')

    except Exception as ex:
        # print(ex)
        pass

def check_versions_for_update(current: str, latest: str):
    current_major, current_minor, *rest = current.split('.')
    latest_major, latest_minor, *rest = latest.split('.')
    if int(current_major) < int(latest_major):
        return True
    if int(current_minor) < int(latest_minor):
        return True
    return False

def sizeof_fmt(num, suffix='B'):
    for unit in ['','Ki','Mi','Gi','Ti']:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, 'Yi', suffix)

def zipdir(path: str, z: zipfile.ZipFile):
    for root, dirs, files in os.walk(path):
        for file in files:
            if not os.path.basename(file).startswith('.'):
                local_path = os.path.join(root, file)
                zip_path = os.path.relpath(os.path.join(root, file), os.path.join(path))
                z.write(local_path, zip_path)



from argparse import ArgumentParser
from pathlib import Path
import os
import shutil

def init_cmd_args(parser: ArgumentParser):
    parser.add_argument(
        "--master_script_slug", "-m", type=str, required=True, help="Slug of the master script"
    )
    parser.add_argument(
        "--task_script_slug", "-t", type=str, required=True, help="Slug of the task script"
    )
    parser.add_argument(
        "--folder_path", "-f", type=str, required=True, help="Destination folder"
    )
    parser.add_argument(
        "--remove", "-r", action='store_true', help="Force removal/overwrite if folder already exists"
    )
    parser.add_argument(
        "--org", "-o", type=str, required=True, help="Org into which the scripts will be uploaded"
    )
    parser.add_argument(
        "--preserve_templates",
        type=bool,
        default=False,
        help="If true, leave template files on disk instead of deleting them",
    )
    parser.set_defaults(func=__cmd)

def __cmd(args):
    # Remove folder if it exists
    new_folder_path = os.path.abspath(args.folder_path)
    new_folder = Path(new_folder_path)
    if new_folder.exists():
        print(f'{new_folder_path} already exists')
        if args.remove:
            shutil.rmtree(new_folder_path)
            print(f'{new_folder_path} removed')
        else:
            print(f'{new_folder_path} exists, do nothing. You can use --remove to remove/overwrite')
            return
    tpl_path = os.path.join(os.path.dirname(__file__), 'protocol-template')
    shutil.copytree(tpl_path, new_folder_path)
    for p in Path(new_folder_path).glob("**/*.template"):
        text = p.read_text()
        text = text.replace("{{ org }}", args.org)
        text = text.replace("{{ master_script_slug }}", args.master_script_slug)
        text = text.replace("{{ task_script_slug }}", args.task_script_slug)

        dest = p.with_name(p.stem)
        print(f'Generating {dest}')
        dest.write_text(text)

        if not args.preserve_templates:
            print(f'Deleting {p}')
            p.unlink()



import os
import sys
import json
import base64
import traceback
import datetime
from collections.abc import Mapping

timestamp_key = "timestamp"
org_slug_key = "orgSlug"
message_key = "message"
file_id_key = "fileId"
master_script_key = "masterScript"
task_script_key = "taskScript"
trace_id_key = "traceId"
container_id_key = "containerId"
level_key = "level"
error_type_key = "errorType"
exception_key = "exception"

class Log:
    def __init__(self, context):
        self._context = context
    
    def log(self, *args):        
        for arg in args:
          try:
            # Perform actual logging, that will be transported to Cloudwatch.
            __builtins__['default_print'](json.dumps(self.create_json_log_entry(arg)))
          except:
            __builtins__['default_print'](arg)
    
    def create_json_log_entry(self, input):
        log_entry = self.generate_default()

        if isinstance(input, str):
            log_entry = {**log_entry, message_key: input, level_key: 'info'}
        elif isinstance(input, Exception):
            log_entry = {**log_entry, **self.generate_error(input)}
        elif isinstance(input, Mapping):
            log_entry = {**log_entry, **input}
        else:
            log_entry = {**log_entry, message_key: input, level_key: 'info'}
        
        if not level_key in log_entry:
            log_entry[level_key] = "info"

        return log_entry

    def generate_default(self):
        try:
            # Lets add only the timestamp right now and add the rest as K8s labels
            
            # master_script_namespace = self._context["masterScriptNamespace"]
            # master_script_slug = self._context["masterScriptSlug"]
            # master_script_version = self._context["masterScriptVersion"]
            # master_script = f"{master_script_namespace}/{master_script_slug}:{master_script_version}"
            
            default_log = {
                timestamp_key: datetime.datetime.now().isoformat(),
                # org_slug_key: self._context["orgSlug"],
                # file_id_key: self._context["inputFile"]["meta"]["fileId"],
                # master_script_key: master_script,
                # task_script_key: self._context["taskScript"]
            }

            if 'CONTAINER_ID' in os.environ:
                default_log[container_id_key] = os.environ.get('CONTAINER_ID')

            if 'traceId' in self._context.get('inputFile', {}).get('meta', {}):
                default_log[trace_id_key] = self._context["inputFile"]["meta"]["traceId"]

            if 'taskId' in self._context:
                default_log['taskId'] = self._context.get('taskId')

            if 'workflowId' in self._context:
                default_log['workflowId'] = self._context.get('workflowId')
        except (AttributeError, KeyError) as e:
            __builtins__['default_print']("Error in default log")
            __builtins__['default_print'](e)
            default_log =  {}

        return default_log
    
    def generate_error(self, err):
        tb_lines = traceback.format_exception(err.__class__, err, err.__traceback__)
        tb_text = ''.join(tb_lines)
        error = { level_key: "error", error_type_key: str(type(err)), exception_key: tb_text }
        if hasattr(err, 'message'):
            error[message_key] = err.message
        else:
            error[message_key] = str(err)
        
        return error



import os
import json
import boto3
import jsonschema
import re

def create_ids_util(locations):

    def get_location(namespace):
        for location in locations:
            if re.search(location['namespacePattern'], namespace):
                return location
        raise Exception(f'Invalid namespace: {namespace}')

    def get_ids(namespace, slug, version):
        location = get_location(namespace)
        bucket = location['bucket']
        prefix = location['prefix']
        endpoint = location['endpoint']
        if endpoint:
            s3 = boto3.client(
                's3', endpoint_url=endpoint,
                aws_access_key_id='123', aws_secret_access_key='abc')
        else:
            s3 = boto3.client('s3')

        file_key = os.path.join(prefix, 'ids', namespace, slug, version, 'schema.json')
        print(f'downloading ids, {bucket}, {file_key}')
        try:
            read_response = s3.get_object(Bucket=bucket, Key=file_key)
            status_code = read_response.get('ResponseMetadata', {}).get('HTTPStatusCode')
            body = read_response.get('Body').read().decode('utf-8')
            return json.loads(body)
        except Exception as e:
            raise Exception(f'Failed to get IDS, {namespace}/{slug}:{version} from bucket: {bucket}, file_key: {file_key}.') from e

    def validate_ids(data, namespace, slug, version):
        ids = get_ids(namespace, slug, version)
        
        validator = jsonschema.Draft7Validator(ids)
        has_error = False
        
        # https://python-jsonschema.readthedocs.io/en/stable/validate/#jsonschema.IValidator.iter_errors
        for error in sorted(validator.iter_errors(data), key=str):
            has_error = True
            print('error: ' + error.message + ' at path: ' + str(list(error.schema_path)))

        if has_error:
            raise Exception(f'Failed validation against IDS: {namespace}/{slug}:{version}')


    return { 'get_ids': get_ids, 'validate_ids': validate_ids }



import json
import os
import io
import zipfile
import argparse
import re
from time import sleep, time

from .__api import upload_artifact, get_task_script_build_info, get_task_script_build_logs, set_ignore_ssl
from .__utils import sizeof_fmt, zipdir

def put_cmd_args(parser: argparse.ArgumentParser):
    parser.add_argument('type', type=str, choices=['ids', 'master-script', 'task-script'])
    parser.add_argument('namespace', type=str)
    parser.add_argument('slug', type=str)
    parser.add_argument('version', type=__version_type)
    parser.add_argument('folder', type=__folder_type, help='path to folder to be uploaded')
    parser.add_argument(
        '--force', '-f', action='store_true', help='force overwrite of an existing artifact'
    )
    parser.add_argument(
        '--ignore-ssl', '-i', action='store_true', help='ignore the SSL certificate verification'
    )
    parser.set_defaults(func=__cmd)

def __version_type(arg_value, pat=re.compile(r'^v')):
    if pat.match(arg_value):
        return arg_value
    return f'v{arg_value}'

def __folder_type(arg_value):
    if os.path.isdir(arg_value):
        return arg_value
    raise argparse.ArgumentTypeError('Not valid folder path provided!')

def __cmd(args):
    set_ignore_ssl(args.ignore_ssl)
    print('Compressing...', flush=True)
    zip_buffer = io.BytesIO()
    zipf = zipfile.ZipFile(zip_buffer, 'a', zipfile.ZIP_DEFLATED, False)
    zipdir(args.folder, zipf)
    zipf.close()
    zip_bytes = zip_buffer.getvalue()

    print(f'Uploading {sizeof_fmt(len(zip_bytes))}...', flush=True)

    r = upload_artifact(args, zip_bytes)

    print(json.dumps(r, indent=4, sort_keys=True), flush=True)

    if args.type == 'task-script':
        build_id = r.get('build', {}).get('id', None)

        if not build_id: 
            raise Exception('No build ID found in upload response!')

        print('Build started', flush=True)
        print("Note: A local script interruption doesn't stop a remote build!", flush=True)

        last_status = None
        prev_next_token = ''

        while True:
            build_info = get_task_script_build_info(build_id)
            build_complete = build_info.get('build', {}).get('buildComplete')
            build_status = build_info.get('build', {}).get('buildStatus')

            sleep(3)

            logs_resp = get_task_script_build_logs(
                build_id,
                {'nextToken': prev_next_token}
            )
            prev_next_token = logs_resp.get('nextToken', None)
            events = logs_resp.get('events', [])

            if len(events) > 0:
                print('\r', end='', flush=True)
            elif not build_complete:
                print('.', end='', flush=True)

            for event in events:
                msg_text = event.get('message', '').strip()
                if msg_text:
                    print(msg_text, flush=True)

            if build_complete:
                last_status = build_status
                break

        print('', flush=True)

        if last_status == 'FAILED':
            raise Exception('Build failed.')



import json
import os
import sys
from main import main
from ts_sdk.taskdev import Context, load_test_file, check_test_output


def test_basic(shared_datadir):
    # set up my for local integration context
    context = Context({ 'mosaic_password': 'secret' })
    test_input_file = load_test_file(
        context, shared_datadir / "input.json", "/input.json", file_category="RAW"
    )
    test_input = {
        'input_file': test_input_file,
        'username': 'username',
        'passwordKey': 'mosaic_password'
    }

    # do the conversion
    actual_output = main(test_input, context)

    # do the check
    check_test_output(
        context, shared_datadir / "expected.json", actual_output, json.loads
    )



import os
import json
import time
import urllib.request
import urllib.parse
import urllib.error
import ssl
import traceback

from datetime import datetime

url = os.environ.get('ORCHESTRATOR_ENDPOINT')
platform_props_hash = os.environ.get('PLATFORM_PROPS_HASH')
task_group_hash = os.environ.get('TASK_GROUP_HASH')
container_id = os.environ.get('CONTAINER_ID')

def poll_task():
  try:
    data = json.dumps({
      'platformPropsHash' : platform_props_hash,
      'taskGroupHash' : task_group_hash,
      'containerId' : container_id
    }).encode('utf8')
    poll_url = url + '/task/poll'
    req = urllib.request.Request(poll_url, data, headers={'content-type': 'application/json'})
    response = urllib.request.urlopen(req, context=ssl._create_unverified_context())
    return generate_task_from_reponse(json.load(response))
  except urllib.error.HTTPError as e:
      print(f'HTTPError: {e.code} for {poll_url}')
  except urllib.error.URLError as e:
      print(f'URLError: {e.code} for {poll_url}')
  except Exception as e:
      print(traceback.format_exc())

def update_task_status(task, result):
  try: 
    data = json.dumps(result).encode('utf8')
    task_id = task.get('id')
    update_url = url + f'/task/{task_id}/update-status'
    req = urllib.request.Request(update_url, data, headers={'content-type': 'application/json'})
    urllib.request.urlopen(req, context=ssl._create_unverified_context())
  except urllib.error.HTTPError as e:
      print(f'HTTPError: {e.code} for {update_url}')
  except urllib.error.URLError as e:
      print(f'URLError: {e.code} for {update_url}')
  except Exception as e:
      print(traceback.format_exc())

def generate_task_from_reponse(body):
  if body:
    data = body.get('data')
    return {
      'id': body.get('id'),
      'context': data.get('context', {}) or {},
      'input': data.get('input', {}) or {},
      'secrets': data.get('secrets', {}) or {},
      'func': data.get('func'),
      'workflow_id': data.get('workflowId'),
      'correlation_id': body.get('correlationId') or body.get('id'),
      'func_dir': data.get('funcDir', './func') or './func'
    }
  
  return {}

def extend_task_timeout(task):
  try:
    task_id = task.get('id')
    extend_timeout_url = url + f'/task/{task_id}/extend-timeout'
    req = urllib.request.Request(extend_timeout_url, {}, headers={'content-type': 'application/json'})
    urllib.request.urlopen(req, context=ssl._create_unverified_context())
    print("EXTENDING: " + task.get('id'))
  except urllib.error.HTTPError as e:
      print(f'HTTPError: {e.code} for {extend_timeout_url}')
      if e.code == 409:
        raise
  except urllib.error.URLError as e:
      print(f'URLError: {e.code} for {extend_timeout_url}')
  except Exception as e:
      print(traceback.format_exc())



# The following imports are necessary for type-checking and editor integration.
from __future__ import annotations
import typing as t

if t.TYPE_CHECKING:
    import ts_sdk.task as ts
# End imports related to type-checking and editor integration.

import io
import json


def convert(input: t.BinaryIO) -> t.Any:
    """Given a file-like object "input", produce a JSON-serializable object
    (nested dictionary, array, etc) that will get dumped to the data lake as
    the output of this workflow.

    ADD YOUR BUSINESS LOGIC HERE.
    """
    return {"a": "1", "label": "label"}


def decorate(input_dict, label):
    input_dict["label"] = label
    return input_dict


def main(input, context: ts.Context):

    IDSNAMESPACE = "common"
    IDSTYPE = "example"
    IDSVERSION = "v1.0.0"

    username = input.get("username")
    passwordKey = input.get("passwordKey")
    password = context.get_secret_config_value(passwordKey)

    ## Add your logic
    file = context.read_file(input["input_file"])
    input_dict = json.loads(file["body"])

    decorate(input_dict, "label")

    convert(io.BytesIO(file["body"]))

    # Validate against IDS
    context.validate_ids(
        input_dict, 
        IDSNAMESPACE, 
        IDSTYPE, 
        IDSVERSION
    )

    # write the IDS JSON to the data lake
    output = context.write_file(
        content=json.dumps(input_dict, indent=2, allow_nan=False),
        file_name="0.json",
        file_category="IDS",
        ids=f"{IDSNAMESPACE}/{IDSTYPE}:{IDSVERSION}",
    )

    print(output)
    return output



# coding: utf-8

import os
from setuptools import setup, find_packages

REQUIRES = [
    "attrs>=20.2.0,<21",
    "boto3>=1.16.3,<2",
    "botocore>=1.19.3,<2",
    "certifi>=2020.6.20,<2021",
    "chardet>=3.0.4,<4",
    "idna>=2.10,<3",
    "importlib-metadata>=2.0.0,<3",
    "jmespath>=0.10.0,<1",
    "jsonschema>=3.2.0,<4",
    "pyrsistent>=0.17.3,<1",
    "python-dateutil>=2.8.1,<3",
    "query-string>=2020.7.1,<2021",
    "requests>=2.24.0,<3",
    "s3transfer>=0.3.3,<1",
    "six>=1.15.0,<2",
    "smart-open[s3]>=3.0.0,<4",
    "typing-extensions>=3.7.4,<4",
    "urllib3>=1.25.11,<2",
    "zipp>=3.3.1,<4"
    ]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

if 'CI' in os.environ and not 'TRAVIS_TAG' in os.environ:
    version = f'1.0.dev{os.environ.get("TRAVIS_BUILD_ID")}'
else:
    version = os.environ.get('TRAVIS_TAG', 'local.dev')

setup(
    name="ts-sdk",
    version=version,
    description="Tetrascience Python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="tetrascience",
    author_email="developers@tetrascience.com",
    url="https://developers.tetrascience.com",
    keywords=[],
    install_requires=REQUIRES,
    packages=find_packages(),
    package_data={'': ['*.txt', '*.json', '*.template', '*.md', '*.py']},
    include_package_data=True,
    python_requires='>=3.7',
    entry_points={
        'console_scripts': {
            'ts-sdk = ts_sdk.cli.__main__:main'
        }
    },
    license='Apache License 2.0'
)



import logging
import boto3
import json

class Storage:
    def __init__(self, bucket, file_key, endpoint):
        self.bucket = bucket
        self.file_key = file_key
        if endpoint:
            self.s3 = boto3.client(
                's3', endpoint_url=endpoint,
                aws_access_key_id='123', aws_secret_access_key='abc')
        else:
            self.s3 = boto3.client('s3')

    def writeObject(self, obj):
        full_file_key = self.file_key.format(**obj)
        response = self.s3.put_object(
            Bucket=self.bucket, Key=full_file_key, Body=json.dumps(obj))
        logging.info(response)



import json
import requests
from datetime import datetime, timezone, timedelta
import time

class Command:

    def __init__(self, endpoint):
        self.endpoint = endpoint

    def run_command(self, context_data, org_slug, target_id, action, metadata, payload, ttl_sec):
        if org_slug is None:
            raise Exception('Param org_slug is missing')
        if target_id is None:
            raise Exception('Param target_id is missing')
        if action is None:
            raise Exception('Param action is missing')
        if payload is None:
            raise Exception('Param payload is missing')
        if ttl_sec < 300 or ttl_sec > 900:
            raise Exception('Param ttl_sec must be between 300 and 900 seconds')

        if metadata is None:
            metadata = {}

        metadata["workflowId"] = context_data.get("workflowId")
        metadata["pipelineId"] = context_data.get("pipelineId")
        metadata["taskId"] = context_data.get("taskId")

        url = self.endpoint + "/internal"
        date_now = datetime.now(timezone.utc) + timedelta(0, ttl_sec)

        command_create_payload = {
            "targetId": target_id,
            "action": action,
            "metadata": metadata,
            "expiresAt": date_now.isoformat(),
            "payload": payload
        }

        headers = { "x-org-slug": org_slug, "Content-Type": "application/json" }

        response = requests.request("POST", url, headers=headers, data=json.dumps(command_create_payload), verify=False)
        if response.status_code == 200:
            print("Command successfully created")
            r = json.loads(response.text)
            command_id = r.get("id")

            command_url = self.endpoint + "/internal/" + command_id
            command_headers = { "x-org-slug": org_slug }

            time_elapsed = 0
            while time_elapsed <= ttl_sec:
                time.sleep(5)
                time_elapsed += 5
                command_response = requests.request("GET", command_url, headers=command_headers, verify=False)
                print("Polling for command status")
                if command_response.status_code == 200:
                    command = json.loads(command_response.text)
                    command_status = command.get("status")
                    print('Current command status: ' + command_status)
                    if command_status == "SUCCESS":
                        return command.get("responseBody")
                    elif command_status == "CREATED" or command_status == "PENDING" or command_status == "PROCESSING":
                        continue
                    else:
                        raise Exception("Command status is " + command_status)

            if time_elapsed >= ttl_sec:
                print('TTL for command has expired')
                raise Exception("Command TTL has expired")
        else:
            print('Error creating command: ' + response.text)
            raise Exception(response.text)


        return response




import io
import typing as t
import typing_extensions as te

FileCategory = te.Literal["IDS", "RAW", "PROCESSED", "TMP"]

class File(te.TypedDict, total=False):
    type: te.Literal["s3"]
    bucket: str
    fileKey: str
    version: t.Optional[str]

class ReadResult(te.TypedDict):
    metadata: t.Mapping[str, str]
    body: t.Optional[bytes]
    file_obj: t.Optional[io.BufferedIOBase]
    download: t.Optional[str]
    custom_metadata: t.Mapping[str, str]
    custom_tags: t.List[str]
